"""Initializer routine for the CODLIR board"""
import logging
import time

from naludaq.backend.managers.io import BoardIoManager
from naludaq.board.initializers import Initializers
from naludaq.helpers.decorators import log_function
from naludaq.communication.spi import SPIConnection, SPIBUS_t
from naludaq.devices.ad5674r import AD5674R_SPI

LOGGER = logging.getLogger("naludaq.init_codlir")

VALID_BOARDS = ["codlir", "codlir_eval"]

DAC_1_map = {
    1: 12,
    3: 13,
    5: 14,
    7: 15,
    9: 4,
    11: 5,
    13: 6,
    15: 7,
    16: 11,
    18: 10,
    20: 9,
    22: 8,
    24: 3,
    26: 2,
    28: 1,
    30: 0,
}
DAC_2_map = {
    0: 12,
    2: 13,
    4: 14,
    6: 15,
    8: 4,
    10: 5,
    12: 6,
    14: 7,
    17: 11,
    19: 10,
    21: 9,
    23: 8,
    25: 3,
    27: 2,
    29: 1,
    31: 0,
}

DAC_3_map = {
    32: 11,
    34: 10,
    36: 9,
    38: 8,
    40: 3,
    42: 2,
    44: 1,
    46: 0,
    49: 12,
    51: 13,
    53: 14,
    55: 15,
    57: 4,
    59: 5,
    61: 6,
    63: 7,
}

DAC_4_map = {
    33: 11,
    35: 10,
    37: 9,
    39: 8,
    41: 3,
    43: 2,
    45: 1,
    47: 0,
    48: 12,
    50: 13,
    52: 14,
    54: 15,
    56: 4,
    58: 5,
    60: 6,
    62: 7,
}


class InitCODLIR(Initializers):
    def __init__(self, board):
        """Initializer for the aardvarcv3.

        Args:
            board (Board): the board to initialize.
        """
        super().__init__(board, VALID_BOARDS)
        self.power_sequence = []

    def run(self, init_clock: bool = True) -> bool:
        """Run the init sequence"""
        has_dacs = True if self.board.features["ext_dac"] else False

        # 1. Write control registers
        self._fpga_init()
        self._i2c_startup()

        # 2. Power up ASICS
        # self._power_toggle(False)
        # self._power_toggle(True)
        time.sleep(0.25)
        # 3. Toggle reset
        # self._system_reset()
        # Program the clock

        # Program the clock
        if init_clock:
            self._init_clock()

        self._system_reset()
        # ControlRegisters(self.board).write("clkrst", 1)
        # time.sleep(0.01)
        # ControlRegisters(self.board).write("clkrst", 0)
        self._set_serial_mode(True)
        time.sleep(0.1)
        self.set_loopback_enabled(False)

        print("Programming ASIC registers")
        self._analog_startup()
        self._digital_startup()

        self._init_spi_connection()
        if has_dacs:
            self._init_dacs()

        return True

    def _init_spi_connection(self):
        """Initialize the SPI connection for the board."""
        self.board.spi_connection = SPIConnection(self.board)
        self.board.spi_connection.add_bus(
            bus_id=0b00,
            bus_type=SPIBUS_t.DAISY_CHAIN,
        )
        self._populate_spi_connection()
        self.board.spi_connection.init()

    def _populate_spi_connection(self):
        """Populate the SPI connection with the devices.

        The devices are connected according to the schematic:
        320-0065-110

        """

        device_map = {
            0: {
                "device_id": 0,
                "bus_id": 0b00,
                "channel_map": DAC_1_map,
            },
            16: {
                "device_id": 1,
                "bus_id": 0b00,
                "channel_map": DAC_2_map,
            },
            32: {
                "device_id": 3,
                "bus_id": 0b00,
                "channel_map": DAC_3_map,
            },
            48: {
                "device_id": 2,
                "bus_id": 0b00,
                "channel_map": DAC_4_map,
            },
        }

        for channel, values in device_map.items():
            bus_id = values["bus_id"]
            device_id = values["device_id"]
            self.board.spi_connection[bus_id][device_id] = AD5674R_SPI(
                spi_connection=self.board.spi_connection,
                bus_id=values["bus_id"],
                device_id=values["device_id"],
                channelmap=values["channel_map"],
                daisy_chained=True,
            )
            # bus_id=values["bus_id"],
            # device_id=values["device_id"],
            # )

    @log_function(LOGGER)
    def _fpga_init(self):
        """Write all FPGA registers to their default values."""
        self.control_registers.write_all()

    @log_function(LOGGER)
    def _i2c_startup(self):
        """Initialize the I2C interface."""
        self.i2c_registers.write("i2c_en", True)

    @log_function(LOGGER)
    def _power_toggle(self, state):
        """Toggles the power rails defined in `power_sequence` attribute."""
        for register in self.power_sequence:
            self.control_write(register, state)

    @log_function(LOGGER)
    def _system_reset(self):
        """Toggle the sysrst pin high, then low"""
        time.sleep(0.01)
        self.control_write("sysrst", True)
        time.sleep(0.01)
        self.control_write("regclr", True)
        time.sleep(0.01)
        self.control_write("sysrst", False)
        time.sleep(0.01)
        self.control_write("regclr", False)

    @log_function(LOGGER)
    def _init_clock(self):
        """Program the clock with the clockfile from the params"""
        LOGGER.info("Programming clock")
        clockfile = self.board.params["clock_file"]
        self.board.clock.program(clockfile)
        # time.sleep(0.25)
        # self.control_write("clk_oeb", False)

    @log_function(LOGGER)
    def _digital_startup(self):
        """Startup the digital side of the chip by programming all registers."""
        self.digital_registers.write_all()

    @log_function(LOGGER)
    def _analog_startup(self):
        """Start the analog side of the chip by programming all registers."""
        self._set_scvbias()
        self.analog_registers.write_all()
        self._dll_startup()

    def _set_scvbias(self):
        """Set the SCVBIAS value for both sides to default values.

        The SCVBias values are set to the values in NaluConfigs.
        """
        time.sleep(0.05)
        self.analog_write("scvbias_left")
        time.sleep(0.05)
        self.analog_write("scvbias_right")
        time.sleep(0.05)

    @log_function(LOGGER)
    def _dll_startup(self):
        """Starting the delay line on the analog side.

        Sets and unsets vanbuff to get the dll going and
        changes the vadjp values to ensure proper SST duty cycle once locked.
        """

        # DLL startup
        for _side in ["left", "right"]:
            self.analog_write(f"qbias_{_side}", 0)
            time.sleep(0.1)
            self.analog_write(f"vadjn_{_side}")  # Use vadjn from yaml
            time.sleep(0.1)
            self.analog_write(f"vanbuf_{_side}", 2000)
            time.sleep(0.5)
            self.analog_write(f"qbias_{_side}", 2950)
            time.sleep(0.1)
            self.analog_write(f"vanbuf_{_side}", 0)
            time.sleep(0.1)

    @log_function(LOGGER)
    def _init_dacs(self):
        """Program the default values to left and right side DACs"""
        channels = list(range(self.board.channels))
        self.board.ext_bias.set_dacs(channels=channels)

    @log_function(LOGGER)
    def _set_serial_mode(self, enabled: bool):
        """Turn on or off serial mode."""
        self.board.control.set_serial_mode(enabled)

    def set_loopback_enabled(self, enabled: bool):
        """Set whether serial loopback is enabled.

        Loopback can safely be disabled during most of the operations with the board.
        Loopback **must** be disabled when communicating over the serial interface.
        If serial communication with the ASIC is intended then this should run during startup and only be enabled as needed.

        Args:
            enabled (bool): True to enable loopback.

        Raises:
            TypeError if enabled is not a bool.
        """
        if not isinstance(enabled, bool):
            raise TypeError("Argument must be bool")
        OFF = "B0900002"
        ON = "B0900003"
        cmd = ON if enabled else OFF
        self._send_command(cmd)

    def _send_command(self, cmd: str):
        """Sends a raw command to the board, handling backend differences."""
        if self.board.using_new_backend:
            BoardIoManager(self.board).write(cmd)
        else:
            self.board.connection.send(cmd)



from naludaq.parsers.headers.base import HeaderParser


class AARDVARCv3HeaderParser(HeaderParser):
    @staticmethod
    def parse_header(event: dict, raw_data: bytes):
        # nibbles: ETT TTT TWW
        # byte one: ETT
        # byte two: TTT
        # byte three: TWW
        # E is an E
        # T is a trigger time nibbles
        # W is previous final window
        trigger_time_ns = raw_data[2] >> 8  # 2 # 24 bit number
        trigger_time_ns += raw_data[1] << 12  # 1
        trigger_time_ns += (raw_data[0] & 0x0FF) << 24
        event["trigger_time_ns"] = trigger_time_ns
        event["prev_final_window"] = raw_data[2] & 0xFF



"""ASoCv3Parser

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 12-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

This module was previously located under the naludaq.daq.workers
subpackage, and was relocated to improve the organization of the
naludaq package.
"""

import logging
from typing import List

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger("naludaq.hdsoc_parser")


class HDSoCParser(Parser):
    def __init__(self, params):
        super().__init__(params)
        self._stop_word = params.get("stop_word", b"\xfa\x5a")
        if isinstance(self._stop_word, str):
            self._stop_word = bytes.fromhex(self._stop_word)
        self._chan_mask = params.get("chanmask", 0x3F)
        self._chan_shift = params.get("chanshift", 0)
        self._abs_wind_mask = params.get("abs_wind_mask", 0x3F)
        self._evt_wind_mask = params.get("evt_wind_mask", 0x3F)
        self._evt_wind_shift = params.get("evt_wind_shift", 6)
        self._headers = params.get("headers", 4)
        self._timing_mask = params.get("timing_mask", 0xFFF)
        self._timing_shift = params.get("timing_mask", 12)
        self._packet_size = params.get("packet_size", 72)

    def _validate_input_package(self, in_data):
        """HDSoC: Splits the input package into a list of packets,
        and validates each packet based on a fixed size. Returns
        a list of only valid packets.

        Args:
            in_data (dict): Raw event structure from packager

        Raises:
            BadDataError: If all packets are bad

        Returns:
            split_in_data (list): List of validated packets
        """
        split_in_data = self._split_rawdata_into_packets(in_data["rawdata"])
        for packet in split_in_data:
            if len(packet) != self._packet_size:
                split_in_data.remove(packet)
        if split_in_data == []:
            raise BadDataError("Input package has no valid packets")
        return split_in_data

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse the raw data from the board.

        Since the data packets are constant length, we can extract the data
        in place with matrix operations, speeding up the parsing.

        Args:
            in_data (bytearray): Raw data from the board

        Returns:
            Parsed event as a dict.

        Raises:
            BadDataException if no data is found or if the data contains errors.
        """
        try:
            input_packets = self._validate_input_package(in_data)
        except:
            raise

        num_packets = len(input_packets)
        raw_data_in = np.frombuffer(np.array(input_packets).flatten(), dtype=">H")
        raw_data = np.reshape(
            np.array(raw_data_in, dtype="uint16"), (num_packets, self._packet_size // 2)
        )
        abs_wind_mask = self._abs_wind_mask
        evt_wind_mask = self._evt_wind_mask
        evt_wind_shift = self._evt_wind_shift
        chan_shift = self._chan_shift
        chan_mask = self._chan_mask
        timing_mask = self._timing_mask
        timing_shift = self._timing_shift
        headers = self._headers

        curr_event = {
            "window_labels": [[] for _ in range(self.params["channels"])],
            "evt_window_labels": [[] for _ in range(self.params["channels"])],
            "data": [[] for _ in range(self.params["channels"])],
            "timing": [[] for _ in range(self.params["channels"])],
            "time": [[] for _ in range(self.params["channels"])],
        }
        np.arange(0, len(raw_data), self._packet_size // 2)
        chans = (raw_data[:, 0] >> chan_shift) & chan_mask
        abs_wins = raw_data[:, 3] & abs_wind_mask
        evt_wins = (raw_data[:, 3] >> evt_wind_shift) & evt_wind_mask
        p1 = np.uint32(raw_data[:, 1] & timing_mask) << timing_shift
        p2 = np.uint32(raw_data[:, 2] & timing_mask)
        timings = p1 | p2

        for i, packet in enumerate(raw_data):
            # If all data bytes are zero the window comes from a disabled channel.
            if packet[12] == 0:
                if np.all(packet[headers:] == 0):
                    continue

            channel = chans[i]  # (packet[0] >> chan_shift) & chan_mask
            abs_window = abs_wins[i]  # packet[3] & abs_wind_mask
            evt_window = evt_wins[i]  # (packet[3] >> evt_wind_shift) & evt_wind_mask
            timing = timings[i]
            curr_event["window_labels"][channel].append(abs_window)
            curr_event["evt_window_labels"][channel].append(evt_window)
            curr_event["timing"][channel].append(timing)
            curr_event["data"][channel].extend(packet[headers : headers + self.samples])
        curr_event["data"] = [np.array(x) for x in curr_event["data"]]
        return curr_event

    def _split_rawdata_into_packets(self, rawdata: bytearray) -> List[bytearray]:
        """Splits raw event data into packets.

        Args:
            rawdata (bytearray): the raw data to split

        Returns:
            A list of raw packets.
        """
        return rawdata.split(self._stop_word)

    def expand_cyclic_labels(self, row_labels):
        """
        Expand a 1D array of cyclic labels into a new 1D array,
        inserting np.nan wherever a label is skipped.

        Example for max_label=61:
        [3,6,7]   -> [3, nan, nan, 6, 7]
        [61,2,3]  -> [61, nan, nan, 2, 3]
        """
        max_label = self.params["windows"]
        row_labels = np.asarray(row_labels, dtype=np.int16)
        if len(row_labels) == 0:
            return np.array([], dtype=float)

        diffs = np.diff(row_labels)
        diffs = np.where(diffs < 0, diffs + (max_label), diffs)
        missing = (
            diffs - 1
        )  # subtract 1 because going from x to x+diffs includes (diffs-1) "skipped" values
        positions = np.zeros(len(row_labels), dtype=int)
        positions[1:] = np.cumsum(1 + missing)

        final_len = positions[-1] + 1
        out_array = np.full(final_len, np.nan, dtype=float)

        out_array[positions] = row_labels

        return out_array

    def _fix_missing_windows(self, event: dict) -> np.ndarray:
        """Expands the window labels to fill in missing windows with np.nan.

        Args:
            event (dict): The event to fix the missing windows in.

        Returns:
            np.ndarray: A 2D array with the missing windows filled with np.nan
        """
        labels_2d = event["window_labels"]

        expanded_rows = [self.expand_cyclic_labels(row) for row in labels_2d]

        max_len = max(len(r) for r in expanded_rows) if expanded_rows else 0
        result = np.full((len(expanded_rows), max_len), np.nan, dtype=float)

        for i, row_expanded in enumerate(expanded_rows):
            result[i, : len(row_expanded)] = row_expanded
        return result

    def expand_cyclic_data(self, row_data, row_labels):
        """
        Expand a 1D array of cyclic data into a new 1D array,
        inserting np.nan wherever a label is skipped.
        """
        samples = self.params["samples"]
        max_label = self.params["windows"]
        row_data = np.asarray(row_data)
        row_labels = np.asarray(row_labels, dtype=int)
        if len(row_data) == 0:
            return np.array([], dtype=float), np.array([], dtype=float)
        diffs = np.diff(row_labels)
        diffs = np.where(diffs < 0, diffs + (max_label), diffs)
        missing = (
            diffs - 1
        )  # subtract 1 because going from x to x+diffs includes (diffs-1) "skipped" values
        positions = np.zeros(len(row_labels), dtype=int)
        positions[1:] = np.cumsum(1 + missing)
        final_len = positions[-1] + 1
        data_array = np.full(final_len * samples, np.nan, dtype=float)
        time_array = np.full(final_len * samples, np.nan, dtype=float)
        datapositions = (positions[:, None] * samples + np.arange(samples)).ravel()
        data_array[datapositions] = row_data
        time_array[datapositions] = datapositions

        return data_array, time_array

    def _fix_missing_data(self, event: dict) -> np.ndarray:
        """Expands the data to fill in missing windows with np.nan.

        Args:
            event (dict): The event to fix the missing windows in.

        Returns:
            np.ndarray: A 2D array with the missing windows filled with np.nan
        """
        labels_2d = event["window_labels"]
        data_2d = event["data"]
        expanded_data = []
        expanded_time = []
        for row_data, row_labels in zip(data_2d, labels_2d):
            expanded_row, time_array = self.expand_cyclic_data(row_data, row_labels)
            expanded_data.append(expanded_row)
            expanded_time.append(time_array)

        max_len = max(len(r) for r in expanded_data) if expanded_data else 0
        data_axis = np.full((len(expanded_data), max_len), np.nan, dtype=float)
        time_axis = np.full((len(expanded_data), max_len), np.nan, dtype=float)

        for i, (data_expanded, time_expanded) in enumerate(
            zip(expanded_data, expanded_time)
        ):
            data_axis[i, : len(data_expanded)] = data_expanded
            time_axis[i, : len(time_expanded)] = time_expanded

        return time_axis, data_axis

    def _add_xaxis_to_event(self, event):
        """Adds an x-axis to the data.

        Based on the amount of channels and samples it will add a timeaxis
        per channel to the event dict.

        During certain readout modes the window labels are not aligning,
        This function accounts for that and moves the time axis accordingly.

        It uses the window labels to determine the time axis by finding the lowest
        window number and offsetting the time axis by that amount.
        The 0-time will be the first samples in the lowest window.

        Args:
            event (dict): The event to add the x-axis to.

        Returns:
            numpy array with sample numbers for each channel.
        """
        channels = self.channels
        maxlen = max([len(x) for x in event["data"]])
        xax = np.tile(np.arange(0, maxlen), channels).reshape(channels, maxlen)

        # Fill in data with np.nan if the window labels are not in order.
        # This operation is channel by channel since the first window label
        # is not necessarily the same for all channels.

        return xax

    def parse(self, raw_data, *args, **kwargs) -> dict:
        """Parses raw_data into a dict.

        Parses the data from the raw event and strips the timestamp/id.
        An dict is created from the parsed data with the id preserved.

        Args:
            raw_data (bytes): Raw data packaged with a header and a stopword.
        Returns:
            dict.
        """
        event = {}
        try:
            event = self.parse_digital_data(raw_data)
        except (TypeError, Exception) as e_msg:
            LOGGER.exception("parse_digital_data failed: %s", e_msg)
            event = raw_data
        else:
            event["time"], event["data"] = self._fix_missing_data(event)
            event["window_labels"] = self._fix_missing_windows(event)
            event["created_at"] = raw_data.get("created_at", 0)
            event["pkg_num"] = raw_data.get("pkg_num", 0)
        event["event_num"] = raw_data.get("pkg_num", 0)
        event["name"] = raw_data.get("name", None)

        return event



"""
"""
from logging import getLogger
from naludaq.controllers.trigger.hdsocv2 import TriggerControllerHdsocv2

logger = getLogger("naludaq.trigger_controller_CODLIR_Eval")


class TriggerControllerCodlir(TriggerControllerHdsocv2):
    def __init__(self, board):
        super().__init__(board)

    def _set_trigger_thresholds(self, values: dict[int, int]):
        """Set the trigger values for the individual channels using the values
        stored in the board object.
        """
        return
        values = values or self.values
        self._validate_thresholds_or_raise(values)
        self._set_tsel_enabled()
        for channel, threshold_value in values.items():
            register_name = self._trig_thresh_regname.format(channel)
            if (
                self.board.registers["analog_registers"][register_name]["value"][0]
                != threshold_value
            ):
                self._write_analog_register(register_name, threshold_value)



from .hdsocv2 import DACControllerHDSoCv2


class DACControllerCodlir(DACControllerHDSoCv2):
    def __init__(self, board):
        """Codlir DAC controller.

        Codlir uses Daisy Chained SPI DAC chips AD5674 in two chains
        Due to this configuration the DAC controller works a bit different.
        However the API should be the same as all other DAC controllers.

        Args:
            board: the board object must contain the SPI daisy chains.
        """
        self.board = board

    def _bus_amount_check_or_raise(self):
        if not len(self.spi) >= 1:
            raise AttributeError(
                f"SPI connection should have one bus for the DACs got {len(self.spi)}"
            )

    def _bus_length_check_or_raise(self):
        if len(self.spi[0b00]) != 4:
            raise AttributeError("Bus 0 should have four devices")

    def _get_device_num(self, channel):
        """Return (device_id, bus_id) for a channel.

        This function will compute the correct bus and device number.
        The devices are connected on the Codlir eval board in the following way:
        - 0-15: DAC 0
        - 16-31: DAC 1
        - 32-47: DAC 3
        - 48-63: DAC 2

        Args:
            channel: The channel number.

        Returns:
            Tuple of (device_id, bus_id)

        Raises:
        TypeError: If the channel is not an int.
            ValueError: If the channel is not in the range of 0-63.
        """
        self.validate_channel_or_raise(self.board, channel)
        for spi_bus in self.spi:
            for spi_dev in spi_bus:
                if channel in spi_dev.channelmap.keys():
                    return (spi_dev.bus_id, spi_dev.device_id)
        raise ValueError(f"Channel {channel} not found in SPI connection")



"""ASoCv3Parser

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 12-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

This module was previously located under the naludaq.daq.workers
subpackage, and was relocated to improve the organization of the
naludaq package.
"""

import logging

import numpy as np

from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger(__name__)


class ASoCv3Parser(Parser):
    def __init__(self, params):
        super().__init__(params)

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse the raw data from the board.
        Assuming that the data packets are constant length, we can extract the data
        in place with matrix operations, speeding up the parsing.

        Parses the data spit out by the digitial readout portion of the firmware

        Channel data is interleaved, e.g
        pkt0,4,8,.... - ch0,
        pkt1,5,9,.... - ch1,
        etc

        TODO: To speed this up more, we do not want two matrices (bitshift & raw),
        we would perferably only want one.

        Args:
            in_data (bytearray): Raw data from the board
        Returns:
            Parsed event as a dict.
        Raises:
            BadDataException if no data is found or if the data contains errors.

        Additional info:
        The word is 16-bit and the data is 12-bit.
        The data is in the form:
        ZZZX XXXX XXXX XXXZ
        Where Z are unused bits.
        0x00 is not a valid byte, ever

        Example format for ASoC:
        first three bits are the "identifier", last bit is always 1
        4 x event headers:
        '001' 'prev_final_window' '1'
        '001' 'trigger_time_ns(23 dt 12)' '1'
        '001' 'trigger_time_ns(11 dt 0 )' '1'
        '001' 'eventNumber' '1'
        1x window header:
        '010' 'chan' '00' 'window' '1'
        64x data packets
        '100' 'data' '1'
        1x end packet (x"FACE")

        """
        return self._parse_event(in_data)
    
    def _parse_event(self, in_data) -> dict:
        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")
        num_headers = self.params.get("headers", 3)
        wlc_stride: "int | None" = self._get_wlc_stride(in_data["rawdata"])
        is_wlc_event: bool = wlc_stride is not None
        # The raw_data is in the form:
        # ZZZX XXXX XXXX XXXZ
        # Where Z are unused bits.

        # Bitshifted data will make it in the form:
        # 0000 XXXX XXXX XXXX
        raw_event_matrix = self._reshape_data(raw_data, is_wlc_event, wlc_stride)
        
        packet_size = self.num_wind_header + self.params["samples"]
        # Pull channel info from all packets
        channels = (raw_event_matrix[0, :] & self.params["chanmask"]) >> self.params[
            "chanshift"
        ]
        available_channels = np.unique(channels[: self.params["channels"]])
        num_windows = raw_event_matrix.shape[1] // len(available_channels)
        curr_event = self._get_event_template(num_windows)

        # Parse headers
        self._parse_event_headers(curr_event, self._bitshift_data(raw_data[:num_headers]))
        first_window_header_idx = num_headers + 1 if is_wlc_event else num_headers
        curr_event["start_window"] = np.uint16(
            (raw_data[first_window_header_idx] & self.params["windmask"])
            >> self.last_bits
        )

        # Parse event data
        for idx, chan in enumerate(available_channels):
            curr_event["data"][chan] = (
            (
            raw_event_matrix[1:packet_size, idx :: len(available_channels)].flatten(order="F") 
            ) & self.data_bitmask
            ) >> self.last_bits
            curr_event["window_labels"][chan] = (
                (
                    raw_event_matrix[0, idx :: len(available_channels)]
                    & self.params["windmask"]
                )
                >> self.last_bits
            )

        if is_wlc_event:
            wlc_headers = (
                raw_data[num_headers:-self.num_footers:wlc_stride]
                & self.params.get("wlcmask", 510)
            ) >> self.last_bits
            curr_event["wlc_headers"] = wlc_headers

        return curr_event


    def _get_event_template(self, num_windows: int) -> dict:
        """Returns a template for the event dictionary."""
        data = np.full((self.params["channels"], num_windows * self.params["samples"]), fill_value=np.nan, dtype=np.float32)
        window_labels = np.full((self.params["channels"], num_windows), fill_value=np.nan, dtype=np.float16)
        event_template = {
            "window_labels": window_labels,
            "data": data,
        }
        return event_template
    
    def _reshape_data(self, data: np.ndarray, is_wlc_event=False, wlc_stride=None) -> np.ndarray:
        """
        Reshapes the data into an np.array of shape (packet_size, num_packets).
        Will ignore headers, footers, and wlc_headers.
        """
        headers = self.params.get("headers", 3)
        data_end = len(data) - self.num_footers
        packet_size = self.num_wind_header + self.params["samples"]
        if is_wlc_event:
            num_channels = (wlc_stride - 1) // packet_size
            num_windows = (len(data) - headers - self.num_footers + 1) // wlc_stride
            num_packets = num_windows * num_channels
            data_by_wlc_blocks = np.reshape(
                data[headers:-1],
                [num_windows, wlc_stride],
            )
            reshaped_data = np.reshape(
                data_by_wlc_blocks[:, 1:].flatten(),
                [packet_size, num_packets],
                order="F",
            )
        else:
            event_size = len(data) - self.num_footers - headers + 1
            num_packets = event_size // packet_size
            reshaped_data = np.reshape(
                data[headers : data_end + 1],
                [packet_size, num_packets],
                order="F",
            )
        
        return reshaped_data

    def _get_wlc_stride(self, raw_data: bytearray) -> "int | None":
        """Checks the length between the first and second occurrence of 0x9E."""
        possible_strides = [(self.params["samples"] + self.num_wind_header) * (chan + 1) + 1 for chan in range(self.params["channels"])]
        try:
            start = raw_data[::2].index(0x9E)
            for possible_stride in possible_strides:
                if raw_data[2 * (start + possible_stride)] == 0x9E:
                    stride = possible_stride
                    break
            else:
                stride = None
        except ValueError:
            stride = None
        return stride

    def _add_xaxis_to_event(self, event):
        """Adds an x-axis to the data.

        Based on the amount of channels and samples it will add a timeaxis
        per channel to the event dict.

        Returns:
            numpy array with sample numbers for each channel.
        """
        if "wlc_headers" in event.keys():
            samples = self.samples
            channels = self.channels
            tim = (samples * np.repeat(event["wlc_headers"], samples)) + (
                np.ones((len(event["wlc_headers"]), samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()
            times = np.tile(tim, (channels, 1))
        else:
            times = super()._add_xaxis_to_event(event)
        return times